package net.wanji.web.service.impl;

import cn.hutool.http.HttpUtil;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import net.wanji.common.utils.geo.CalUtils;
import net.wanji.common.utils.tool.CrossUtil;
import net.wanji.databus.dao.entity.RidInfoEntity;
import net.wanji.databus.dao.mapper.BaseCrossInfoMapper;
import net.wanji.databus.dao.mapper.RidInfoMapper;
import net.wanji.databus.po.BaseCrossInfoPO;
import net.wanji.databus.po.TBaseCrossInfo;
import net.wanji.web.bo.SpecialServiceRouteBO;
import net.wanji.web.common.enums.CrossDirEnum;
import net.wanji.web.common.enums.CrossPhasePlanTurnTypeEnum;
import net.wanji.web.common.enums.SpecialServiceCrossTurnEnum;
import net.wanji.web.common.exception.CrossRelationException;
import net.wanji.web.mapper.*;
import net.wanji.web.po.CrossDirTurnPO;
import net.wanji.web.po.RidInfoPO;
import net.wanji.web.po.SpecialServiceCrossPO;
import net.wanji.web.po.SpecialServicePO;
import net.wanji.web.service.SpecialServiceService;
import net.wanji.web.vo.specialService.*;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.*;

/**
 * @author Kent HAN
 * @date 2022/11/8 11:07
 */
@Service
public class SpecialServiceServiceImpl implements SpecialServiceService {
    // key所有路段的多边形
    Map<double[][], RidInfoEntity> allRidPolygons = new HashMap<>();


    // 所有路段，key起始点，value路段实体类
    Map<double[][], RidInfoEntity> allStartEnds = new HashMap<>();
    // 相交路段
    List<RidInfoEntity> straddleRids = new ArrayList<>();

    @Autowired
    SpecialServiceMapper specialServiceMapper;
    @Autowired
    SpecialServiceCrossMapper specialServiceCrossMapper;
    @Autowired
    CustomRidInfoMapper customRidInfoMapper;
    @Autowired
    BaseCrossInfoMapper tBaseBaseCrossInfoMapper;
    @Autowired
    CrossDirTurnMapper crossDirTurnMapper;
    @Autowired
    CrossPhasePlanMapper crossPhasePlanMapper;
    @Autowired
    private BaseCrossInfoMapper baseCrossInfoMapper;
    @Autowired
    private RidInfoMapper ridInfoMapper;

    private Gson gson = new GsonBuilder().setDateFormat("yyyy-MM-dd HH:mm:ss").create();

    @Override
    public void addSpecialService(AddSpecialServiceInVO addSpecialServiceInVO) {
        SpecialServicePO specialServicePO = new SpecialServicePO();
        String name = addSpecialServiceInVO.getName();
        specialServicePO.setName(name);
        specialServiceMapper.insertOne(specialServicePO);
    }

    @Override
    public List<ListSpecialServicesOutVO> listSpecialServices(String adCode) {
        List<SpecialServicePO> specialServicePOList = specialServiceMapper.selectByAdCode(adCode);
        List<ListSpecialServicesOutVO> listSpecialServicesOutVOList = new ArrayList<>();
        for (SpecialServicePO specialServicePO : specialServicePOList) {
            ListSpecialServicesOutVO listSpecialServicesOutVO = new ListSpecialServicesOutVO();
            BeanUtils.copyProperties(specialServicePO, listSpecialServicesOutVO);
            // 构造wkt列表
            List<String> wkts = new ArrayList<>();
            Integer specialServiceId = specialServicePO.getId();
            List<SpecialServiceCrossPO> specialServiceCrossPOList =
                    specialServiceCrossMapper.selectBySpecialServiceId(specialServiceId);
            int size = specialServiceCrossPOList.size();
            for (int i = 0; i < size - 1; i++) {
                SpecialServiceCrossPO currentCross = specialServiceCrossPOList.get(i);
                SpecialServiceCrossPO nextCross = specialServiceCrossPOList.get(i + 1);
                RidInfoPO ridInfoPO = customRidInfoMapper.selectOne(currentCross.getCrossId(), nextCross.getCrossId());
                String wkt = ridInfoPO.getWkt();
                wkts.add(wkt);
            }
            listSpecialServicesOutVO.setWkts(wkts);
            listSpecialServicesOutVOList.add(listSpecialServicesOutVO);
        }
        return listSpecialServicesOutVOList;
    }

    @Override
    public void updateSpecialService(UpdateSpecialServiceInVO updateSpecialServiceInVO) {
        specialServiceMapper.updateSpecialService(updateSpecialServiceInVO);
    }

    @Override
    @Transactional
    public void deleteSpecialService(Integer id) {
        specialServiceMapper.deleteSpecialService(id);
        specialServiceCrossMapper.deleteCrossBySpecialServiceId(id);
    }

    @Override
    @Transactional
    public void addCross(AddCrossInVO addCrossInVO) throws RuntimeException{
        Integer specialServiceId = addCrossInVO.getSpecialServiceId();
        String endCrossid = addCrossInVO.getCrossId();
        SpecialServiceCrossPO specialServiceCrossPO = new SpecialServiceCrossPO();
        BeanUtils.copyProperties(addCrossInVO, specialServiceCrossPO);
        // 构造顺序号
        Integer maxSort = specialServiceCrossMapper.getMaxSort(specialServiceId);
        if (maxSort == null) {
            maxSort = 1;
        } else {
            maxSort++;
        }
        if (maxSort > 1) {
            // 插入路段长度，插入转向
            // 查询上一个路口ID
            Integer preSort = maxSort - 1;
            String startCrossid = specialServiceCrossMapper.selectCrossIdBySort(specialServiceId, preSort);
            // 查询路段
            RidInfoPO ridInfoPO = customRidInfoMapper.selectOne(startCrossid, endCrossid);
            // 路段关系不正确
            if (ridInfoPO == null) {
                throw new CrossRelationException("路段关系不正确");
            }
            // 插入路段长度
            Double roadLength = ridInfoPO.getLength();
            Double oldLength = specialServiceMapper.selectRoadLength(specialServiceId);
            oldLength += roadLength;
            specialServiceMapper.updateLength(specialServiceId, oldLength);
            // 插入上一个路口转向
            String startCrossOutDirId = ridInfoPO.getStartCrossdirid();
            Integer startCrossInDir = specialServiceCrossMapper.selectStartCrossInDir(specialServiceId, startCrossid);
            String startCrossInDirId = startCrossid + "_" + startCrossInDir + "_2_1";
            CrossDirTurnPO crossDirTurnPO = crossDirTurnMapper.selectOneByInAndOutIds(startCrossInDirId, startCrossOutDirId);
            String turntype = crossDirTurnPO.getTurntype();
            specialServiceCrossMapper.updateTurn(specialServiceId, startCrossid, turntype);
            // 更新当前路口进口
            String endDirCode = ridInfoPO.getEndDir();
            Integer endDirInt = Integer.parseInt(endDirCode);
            specialServiceCrossPO.setDir(endDirInt);
        }
        specialServiceCrossPO.setSort(maxSort);
        specialServiceCrossMapper.insertOne(specialServiceCrossPO);
    }

    @Override
    public List<CrossListOutVO> crossList(Integer specialServiceId) {
        List<SpecialServiceCrossPO> specialServiceCrossPOList =
                specialServiceCrossMapper.selectBySpecialServiceId(specialServiceId);
        List<CrossListOutVO> crossListOutVOList = new ArrayList<>();
        for (int i = 0; i < specialServiceCrossPOList.size(); i++) {
            SpecialServiceCrossPO specialServiceCrossPO = specialServiceCrossPOList.get(i);
            CrossListOutVO crossListOutVO = new CrossListOutVO();
            String crossId = specialServiceCrossPO.getCrossId();
            crossListOutVO.setCrossId(crossId);
            Integer sort = specialServiceCrossPO.getSort();
            crossListOutVO.setSort(sort);
            crossListOutVO.setStatus(specialServiceCrossPO.getStatus());
            TBaseCrossInfo tBaseCrossInfo = tBaseBaseCrossInfoMapper.selectByPrimaryKey(specialServiceCrossPO.getCrossId());
            // 经纬度
            String location = tBaseCrossInfo.getLocation();
            double[] lonLat = CrossUtil.getLonLat(location);
            crossListOutVO.setLongitude(lonLat[0]);
            crossListOutVO.setLatitude(lonLat[1]);
            // 路口名称
            crossListOutVO.setName(tBaseCrossInfo.getName());
            // 获取控制方向
            Integer dirCode = specialServiceCrossPO.getDir();
            String dir = CrossDirEnum.getMsgByCode(dirCode);
            String turnCode = specialServiceCrossPO.getTurn();
            if (turnCode != null) {
                String turn = SpecialServiceCrossTurnEnum.getMsgByCode(turnCode);
                String dirAndTurn = dir + turn;
                crossListOutVO.setDirAndTurn(dirAndTurn);
            }
            // 获取wkt
            if (i < specialServiceCrossPOList.size() - 1) {
                SpecialServiceCrossPO nextPO = specialServiceCrossPOList.get(i + 1);
                String startCrossid = specialServiceCrossPO.getCrossId();
                String endCrossid = nextPO.getCrossId();
                String wkt = customRidInfoMapper.selectWkt(startCrossid, endCrossid);
                crossListOutVO.setWkt(wkt);
            }
            crossListOutVOList.add(crossListOutVO);
        }

        return crossListOutVOList;
    }

    @Override
    @Transactional
    public void updateCross(UpdateCrossInVO updateCrossInVO) {
        Integer specialServiceId = updateCrossInVO.getSpecialServiceId();
        String crossId = updateCrossInVO.getCrossId();
        String dirInAndOut = updateCrossInVO.getDirInAndOut();
        // 获取方向（东西南北...）
        int length = dirInAndOut.length();
        String dirStr = dirInAndOut.substring(0, length - 2);
        Integer dir = CrossDirEnum.getCodeByMsg(dirStr);
        specialServiceCrossMapper.updatDir(specialServiceId, crossId, dir);
        // 更新当前路口转向
        Integer sort = updateCrossInVO.getSort();
        Integer nextSort = sort + 1;
        String endCrossid = specialServiceCrossMapper.selectCrossIdBySort(specialServiceId, nextSort);
        String startCrossOutDirId = customRidInfoMapper.selectStartCrossOutDir(crossId, endCrossid);
        String startCrossInDirId = crossId + "_" + dir + "_2_1";
        CrossDirTurnPO crossDirTurnPO = crossDirTurnMapper.selectOneByInAndOutIds(startCrossInDirId, startCrossOutDirId);
        String turntype = crossDirTurnPO.getTurntype();
        specialServiceCrossMapper.updateTurn(specialServiceId, crossId, turntype);
    }

    @Override
    @Transactional
    public void deleteCross(DeleteCrossInVO deleteCrossInVO) {
        String crossId = deleteCrossInVO.getCrossId();
        Integer specialServiceId = deleteCrossInVO.getSpecialServiceId();
        Integer sort = deleteCrossInVO.getSort();
        if (sort > 1) {
            Integer preSort = sort - 1;
            String preCrossid = specialServiceCrossMapper.selectCrossIdBySort(specialServiceId, preSort);
            // 减少路线长度
            Double roadLength = customRidInfoMapper.selectRoadLength(preCrossid, crossId);
            Double oldLength = specialServiceMapper.selectRoadLength(specialServiceId);
            oldLength -= roadLength;
            specialServiceMapper.updateLength(specialServiceId, oldLength);
            // 删除上一个路口转向
            specialServiceCrossMapper.updateTurn(specialServiceId, preCrossid, null);
        }
        specialServiceCrossMapper.deleteCross(specialServiceId, crossId);
    }

    @Override
    public void enableSpecialService(EnableDisableSpecialServiceInVO enableDisableSpecialServiceInVO) {
        Integer specialServiceId = enableDisableSpecialServiceInVO.getSpecialServiceId();
        specialServiceMapper.updateStatusEnable(specialServiceId);
        List<SpecialServiceCrossPO> specialServiceCrossPOList =
                specialServiceCrossMapper.selectBySpecialServiceId(specialServiceId);
        for (SpecialServiceCrossPO specialServiceCrossPO : specialServiceCrossPOList) {
            // 构造请求体
            EnableDisableSpecialServiceOutVO enableDisableSpecialServiceOutVO = new EnableDisableSpecialServiceOutVO();

            PhaseLockVO phaseLockVO = new PhaseLockVO();
            Integer dirInt = specialServiceCrossPO.getDir();
            String dirStr = dirInt.toString();
            String turnStr = specialServiceCrossPO.getTurn();
            String turnCode = CrossPhasePlanTurnTypeEnum.getCodeByMsg(turnStr);
            String phaseId = crossPhasePlanMapper.selectPhaseId(dirStr, turnCode);
            phaseLockVO.setPhaseid(phaseId);
            List<PhaseLockVO> phaseLockVOList = Arrays.asList(phaseLockVO);
            enableDisableSpecialServiceOutVO.setPhases(phaseLockVOList);
            enableDisableSpecialServiceOutVO.setCommand(1);

            String crossId = specialServiceCrossPO.getCrossId();
            // 发送请求
            String url = "http://10.102.1.182/utc/controlCommand/lockControl";
            String s = gson.toJson(enableDisableSpecialServiceOutVO);
            String response = HttpUtil.post(url, s);
            EnableDisableSpecialServiceResponseVO enableDisableSpecialServiceResponseVO =
                    gson.fromJson(response, EnableDisableSpecialServiceResponseVO.class);
            Boolean success = enableDisableSpecialServiceResponseVO.getSuccess();

            // 更新路口状态
            if (success) {
                specialServiceCrossMapper.updateStatusSuccess(specialServiceId, crossId);
            } else {
                specialServiceCrossMapper.updateStatusFail(specialServiceId, crossId);
            }
        }
    }

    @Override
    public void disableSpecialService(EnableDisableSpecialServiceInVO enableDisableSpecialServiceInVO) {
        Integer specialServiceId = enableDisableSpecialServiceInVO.getSpecialServiceId();
        specialServiceMapper.updateStatusDisable(specialServiceId);
        List<SpecialServiceCrossPO> specialServiceCrossPOList =
                specialServiceCrossMapper.selectBySpecialServiceId(specialServiceId);
        for (SpecialServiceCrossPO specialServiceCrossPO : specialServiceCrossPOList) {
            // 构造请求体
            EnableDisableSpecialServiceOutVO enableDisableSpecialServiceOutVO = new EnableDisableSpecialServiceOutVO();

            PhaseLockVO phaseLockVO = new PhaseLockVO();
            Integer dirInt = specialServiceCrossPO.getDir();
            String dirStr = dirInt.toString();
            String turnStr = specialServiceCrossPO.getTurn();
            String turnCode = CrossPhasePlanTurnTypeEnum.getCodeByMsg(turnStr);
            String phaseId = crossPhasePlanMapper.selectPhaseId(dirStr, turnCode);
            phaseLockVO.setPhaseid(phaseId);
            List<PhaseLockVO> phaseLockVOList = Arrays.asList(phaseLockVO);
            enableDisableSpecialServiceOutVO.setPhases(phaseLockVOList);
            enableDisableSpecialServiceOutVO.setCommand(0);

            String crossId = specialServiceCrossPO.getCrossId();

            // 发送请求
            String url = "http://10.102.1.182/utc/controlCommand/lockControl";
            String s = gson.toJson(enableDisableSpecialServiceOutVO);
            String response = HttpUtil.post(url, s);
            EnableDisableSpecialServiceResponseVO enableDisableSpecialServiceResponseVO =
                    gson.fromJson(response, EnableDisableSpecialServiceResponseVO.class);
            Boolean success = enableDisableSpecialServiceResponseVO.getSuccess();

            // 更新路口状态
            if (success) {
                specialServiceCrossMapper.updateStatusSuccess(specialServiceId, crossId);
            } else {
                specialServiceCrossMapper.updateStatusFail(specialServiceId, crossId);
            }
        }
    }

    @Override
    public Set<String> crossInList(String crossId) {
        List<String> crossInIdList = crossDirTurnMapper.crossInList(crossId);
        Set<String> res = new HashSet<>();
        for (String crossInId : crossInIdList) {
            String crossIn = getCrossIn(crossInId);
            res.add(crossIn);
        }
        return res;
    }

    private String getCrossIn(String crossInId) {
        // 从 13FNK0C6790_1_2_1 中获取 “北进口”
        int length = crossInId.length();
        String dir = crossInId.substring(length - 5, length - 4);
        String dirStr = CrossDirEnum.getMsgByCode(Integer.parseInt(dir));
        String s = dirStr + "进口";
        return s;
    }

    @Override
    public void specialServiceRoute(SpecialServiceRouteBO specialServiceRouteBO) {
        // 初始化所有路段多边形
        initAllPolygons();
        // 获取起点路口和终点路口
        double[] startLonLat = specialServiceRouteBO.getStartLonLat();
        double[] endLonLat = specialServiceRouteBO.getEndLonLat();
        double startMinDistance = Double.MAX_VALUE;
        double endMinDistance = Double.MAX_VALUE;
        BaseCrossInfoPO startCross = null;
        BaseCrossInfoPO endCross = null;

        for (double[][] ridPolygon : allRidPolygons.keySet()) {
            boolean isStartInPolygon = CalUtils.isPointinPolygon(startLonLat, ridPolygon);
            boolean isEndInPolygon = CalUtils.isPointinPolygon(endLonLat, ridPolygon);
            if (isStartInPolygon) {
                RidInfoEntity ridInfoEntity = allRidPolygons.get(ridPolygon);

                String ridStartCrossId = ridInfoEntity.getStartCrossId();
                BaseCrossInfoPO ridStartCrossPO = baseCrossInfoMapper.selectById(ridStartCrossId);
                String ridStartLocation = ridStartCrossPO.getLocation();
                double[] ridStartLonLat = CrossUtil.getLonLat(ridStartLocation);
                double distanceToRidStart = CalUtils.calculate_distance(
                        ridStartLonLat[0], ridStartLonLat[1], startLonLat[0], startLonLat[1]);
                startMinDistance = Math.min(startMinDistance, distanceToRidStart);
                if (distanceToRidStart == startMinDistance) {
                    startCross = ridStartCrossPO;
                }

                String ridEndCrossId = ridInfoEntity.getEndCrossId();
                BaseCrossInfoPO ridEndCrossPO = baseCrossInfoMapper.selectById(ridEndCrossId);
                String ridEndLocation = ridEndCrossPO.getLocation();
                double[] ridEndLonLat = CrossUtil.getLonLat(ridEndLocation);
                double distanceToRidEnd = CalUtils.calculate_distance(
                        ridEndLonLat[0], ridEndLonLat[1], startLonLat[0], startLonLat[1]);
                startMinDistance = Math.min(startMinDistance, distanceToRidEnd);
                if (distanceToRidEnd == startMinDistance) {
                    startCross = ridEndCrossPO;
                }
            }

            if (isEndInPolygon) {
                RidInfoEntity ridInfoEntity = allRidPolygons.get(ridPolygon);

                String ridStartCrossId = ridInfoEntity.getStartCrossId();
                BaseCrossInfoPO ridStartCrossPO = baseCrossInfoMapper.selectById(ridStartCrossId);
                String ridStartLocation = ridStartCrossPO.getLocation();
                double[] ridStartLonLat = CrossUtil.getLonLat(ridStartLocation);
                double distanceToRidStart = CalUtils.calculate_distance(
                        ridStartLonLat[0], ridStartLonLat[1], startLonLat[0], startLonLat[1]);
                endMinDistance = Math.min(endMinDistance, distanceToRidStart);
                if (distanceToRidStart == endMinDistance) {
                    endCross = ridStartCrossPO;
                }

                String ridEndCrossId = ridInfoEntity.getEndCrossId();
                BaseCrossInfoPO ridEndCrossPO = baseCrossInfoMapper.selectById(ridEndCrossId);
                String ridEndLocation = ridEndCrossPO.getLocation();
                double[] ridEndLonLat = CrossUtil.getLonLat(ridEndLocation);
                double distanceToRidEnd = CalUtils.calculate_distance(
                        ridEndLonLat[0], ridEndLonLat[1], endLonLat[0], endLonLat[1]);
                endMinDistance = Math.min(endMinDistance, distanceToRidEnd);
                if (distanceToRidEnd == endMinDistance) {
                    endCross = ridEndCrossPO;
                }
            }
        }
        // 返回两条备选线路
        System.out.println("todo");
    }

    private void initAllPolygons() {
        allRidPolygons.clear();
        List<RidInfoEntity> ridInfoEntities = ridInfoMapper.selectAll();
        for (RidInfoEntity ridInfoEntity : ridInfoEntities) {
            String wkt = ridInfoEntity.getWkt();
            String[] lonLats = wkt.split(";");
            int length = lonLats.length;
            double[][] doubles = new double[length + 1][2]; // +1是为了首尾相接，算法工具类需要
            for (int i = 0; i < length; i++) {
                String lonLat = lonLats[i];
                String[] pointLonLat = lonLat.split(",");
                double pointLon = Double.parseDouble(pointLonLat[0]);
                double pointLat = Double.parseDouble(pointLonLat[1]);
                double[] pointLonLatDouble = {pointLon, pointLat};
                doubles[i] = pointLonLatDouble;
                if (i == length - 1) { // 最后一个点位
                    doubles[i + 1] = doubles[0]; // 首尾相接，算法工具类需要
                }
            }
            allRidPolygons.put(doubles, ridInfoEntity);
        }
    }

    private void fillAllStartEnds() {
        allStartEnds.clear();
        List<RidInfoEntity> ridInfoEntities = ridInfoMapper.selectAll();
        for (RidInfoEntity ridInfoEntity : ridInfoEntities) {
            String startCrossId = ridInfoEntity.getStartCrossId();
            String endCrossId = ridInfoEntity.getEndCrossId();

            BaseCrossInfoPO startCrossPO = baseCrossInfoMapper.selectById(startCrossId);
            String startLocation = startCrossPO.getLocation();
            double[] startLonLat = CrossUtil.getLonLat(startLocation);

            BaseCrossInfoPO endCrossPO = baseCrossInfoMapper.selectById(endCrossId);
            String endLocation = endCrossPO.getLocation();
            double[] endLonLat = CrossUtil.getLonLat(endLocation);

            double[][] doubles = new double[][]{startLonLat, endLonLat};
            allStartEnds.put(doubles, ridInfoEntity);
        }
    }
}